import os
import shutil
from unittest.mock import MagicMock

import ansiblecall
import ansiblecall.utils.loader


def test_salt(monkeypatch):
    """Mock __salt__ and test two part reference."""
    has_salt = MagicMock(return_value=True)
    monkeypatch.setattr(ansiblecall.utils.loader, "has_salt", has_salt)
    ansiblecall.refresh_modules()
    has_salt.assert_called()
    ret = ansiblecall.module("ansible_builtin.ping")
    assert ret == {"ping": "pong"}

    # Reset mock
    has_salt = MagicMock(return_value=False)
    monkeypatch.setattr(ansiblecall.utils.loader, "has_salt", has_salt)
    ansiblecall.refresh_modules()


def test_zip(monkeypatch):
    """Check module run from zip file"""
    shutil.rmtree(os.path.expanduser("~/.ansiblecall"))
    ansiblecall.cache(mod_name="ansible.builtin.ping")
    monkeypatch.syspath_prepend(os.path.expanduser("~/.ansiblecall/cache/ansible.builtin.ping.zip"))
    ansiblecall.utils.loader.reload()
    ret = ansiblecall.module("ansible.builtin.ping")
    assert ret == {"ping": "pong"}
    assert ansiblecall.__file__ == os.path.expanduser(
        "~/.ansiblecall/cache/ansible.builtin.ping/ansiblecall/__init__.py"
    )
    shutil.rmtree(os.path.expanduser("~/.ansiblecall"))
    monkeypatch.undo()
    ansiblecall.utils.loader.reload()



import logging
import pathlib
import tempfile

import ansiblecall

log = logging.getLogger(__name__)


def test_escalate_privilege():
    rt_root = ansiblecall.Runtime(become=True)
    ret = ansiblecall.module("ansible.builtin.command", rt=rt_root, argv=["whoami"])
    assert ret["stdout"] == "root"

    # Create a new user
    user = "john"
    ret = ansiblecall.module("ansible.builtin.user", rt=rt_root, name=user, create_home="yes")
    assert ret["state"] == "present"

    # Run a command as this new user
    rt = ansiblecall.Runtime(become=True, become_user=user)
    ret = ansiblecall.module("ansible.builtin.command", rt=rt, argv=["whoami"])
    assert ret["stdout"] == user

    # Delete the user
    # Github runner doesn't allow deleting a user. Comment it for now.
    # ret = ansiblecall.module("ansible.builtin.user", rt=rt_root, name=user, state="absent")
    # assert ret["state"] == "absent"

    # Touch a file as root
    with tempfile.TemporaryDirectory() as tmp_dir:
        foo_file = str(pathlib.Path(tmp_dir).joinpath("foo"))
        foo_archive = str(pathlib.Path(tmp_dir).joinpath("foo.gz"))
        ret = ansiblecall.module(
            mod_name="ansible.builtin.file",
            path=str(foo_file),
            state="touch",
            rt=rt_root,
        )
        assert ret["changed"] is True
        ansiblecall.module(
            mod_name="ansible.builtin.file",
            path=foo_archive,
            state="absent",
            rt=rt_root,
        )
        ret = ansiblecall.module(mod_name="community.general.archive", path=foo_file, rt=rt_root)
        assert ret["changed"] is True



import functools
import glob
import importlib
import logging
import os
import pathlib
import sys
import time

log = logging.getLogger(__name__)


def has_salt():
    return "__salt__" in globals()


def load_module(module_key, module_name, module_path, module_abs):
    # Lazy import
    import ansiblecall

    ret = {}
    proxy_mod = functools.partial(ansiblecall.module, name=module_key)
    proxy_mod.path = module_path
    proxy_mod.name = module_name
    proxy_mod.abs = module_abs
    proxy_mod.key = module_key
    ret[module_key] = proxy_mod
    return ret


@functools.lru_cache
def load_mods():
    """Load ansible modules"""
    # Lazy import
    import ansible.modules

    ret = {}
    # Load ansible core modules
    for path in ansible.modules.__path__:
        if str(pathlib.Path(path).parent.parent).endswith(".zip"):
            continue
        for f in os.listdir(path):
            if f.startswith("_") or not f.endswith(".py"):
                continue
            fname = f.removesuffix(".py")
            # Ansible modules will be referred in salt as 2 parts ansible_builtin.ping instead of
            # ansible.builtin.ping.
            mod = f"ansible_builtin.{fname}" if has_salt() else f"ansible.builtin.{fname}"
            module_name = f"{ansible.modules.__name__}.{fname}"
            module_path = os.path.dirname(os.path.dirname(ansible.__file__))
            ret.update(
                load_module(
                    module_key=mod,
                    module_name=module_name,
                    module_path=module_path,
                    module_abs=os.path.join(*ansible.modules.__path__, f),
                ),
            )

    # Load collections when available
    # Refer: https://docs.ansible.com/ansible/latest/collections_guide/collections_installing.html#installing-collections-with-ansible-galaxy
    roots = sys.path
    roots.append(os.path.expanduser(os.environ.get("ANSIBLE_COLLECTIONS_PATH", "~/.ansible/collections")))
    for collections_root in roots:
        if str(collections_root).endswith(".zip"):
            continue
        # The glob will produce result like below
        # ['/root/.ansible/collections/ansible_collections/amazon/aws/plugins/modules/cloudtrail_info.py', ...]
        for f in glob.glob(os.path.join(collections_root, "ansible_collections/*/*/plugins/modules/*.py")):
            relname = os.path.relpath(f.removesuffix(".py"), collections_root)
            name_parts = relname.split("/")
            namespace, coll_name, module = name_parts[1], name_parts[2], name_parts[-1]
            if module.startswith("_"):
                continue
            # Ansible modules will be referred in salt as 2 parts ansible_builtin.ping instead of
            # ansible.builtin.ping.
            mod = f"{namespace}_{coll_name}.{module}" if has_salt() else f"{namespace}.{coll_name}.{module}"
            module_name = relname.replace("/", ".")
            module_path = collections_root
            module_abs = f
            ret.update(
                load_module(
                    module_key=mod,
                    module_name=module_name,
                    module_path=module_path,
                    module_abs=module_abs,
                ),
            )
    return ret


def finder(fun):
    """
    Find and extract files when a module is imported from zip file
    """

    def wrapped(mod_name, *args, **kwargs):
        import ansiblecall.utils.ctx

        with ansiblecall.utils.ctx.ZipContext(mod_name=mod_name):
            return fun(mod_name, *args, **kwargs)

    return wrapped


@finder
def get_module(mod_name):
    start = time.time()
    modules = load_mods()
    log.debug(
        "Loaded %s ansible modules. Elapsed: %0.03fs",
        len(modules),
        (time.time() - start),
    )
    return modules[mod_name]


def reload():
    import ansible
    import ansible.modules

    import ansiblecall

    for mod in (ansible, ansible.modules, ansiblecall):
        importlib.reload(mod)



import json
import logging
import os
import subprocess
import sys
import tempfile

from ansible.module_utils.common.respawn import _create_payload, has_respawned

from ansiblecall.utils.cache import package_libs

log = logging.getLogger(__name__)


def build_cmd(interpreter_path=None, runtime=None):
    cmd = []
    if runtime:
        if runtime.become:
            cmd.append("sudo")
        if runtime.become_user:
            cmd.extend(["su", runtime.become_user, "-c"])
    cmd.extend([interpreter_path or "python3", "--"])
    return cmd


def own_namespace(fun):
    def wrapped(*args, **kwargs):
        with tempfile.TemporaryDirectory() as tmp_dir:
            package_libs(path=tmp_dir)
            os.chmod(tmp_dir, 0o555)  # noqa: S103
            sys.modules["__main__"]._modlib_path = tmp_dir  # noqa: SLF001
            return fun(*args, **kwargs)

    return wrapped


@own_namespace
def respawn_module(interpreter_path=None, runtime=None):
    """
    Respawn the currently-running Ansible Python module under the specified Python interpreter.

    Ansible modules that require libraries that are typically available only under well-known interpreters
    (eg, ``apt``, ``dnf``) can use bespoke logic to determine the libraries they need are not
    available, then call `respawn_module` to re-execute the current module under a different interpreter
    and exit the current process when the new subprocess has completed. The respawned process inherits only
    stdout/stderr from the current process.

    Only a single respawn is allowed. ``respawn_module`` will fail on nested respawns. Modules are encouraged
    to call `has_respawned()` to defensively guide behavior before calling ``respawn_module``, and to ensure
    that the target interpreter exists, as ``respawn_module`` will not fail gracefully.

    :arg interpreter_path: path to a Python interpreter to respawn the current module
    """

    if has_respawned():
        raise Exception("module has already been respawned")  # noqa: TRY002, TRY003, EM101

    # FUTURE: we need a safe way to log that a respawn has occurred for forensic/debug purposes
    payload = _create_payload()
    # Changes start
    cmd = build_cmd(interpreter_path=interpreter_path, runtime=runtime)
    ret = subprocess.run(
        cmd,
        input=payload,
        text=True,
        capture_output=True,
        check=False,
    )
    out = ret.stdout
    if ret.returncode:
        out = json.dumps({"changed": False, "failed": True, "msg": str(ret.stderr.strip())})
    sys.stdout.flush()
    sys.stdout.write(out)
    # Changes end
    # Cleanup cache files as escalated user
    for target in ["__pycache__", "*.pyc"]:
        cleanup = """
import pathlib
import os
for p in pathlib.Path({modlib_path!r}).rglob({target!r}):
    os.unlink(p)
        """.format(modlib_path=sys.modules["__main__"]._modlib_path, target=target)  # noqa: SLF001
        ret = subprocess.run(
            cmd,
            input=cleanup,
            text=True,
            capture_output=True,
            check=False,
        )

    sys.exit(0)



import importlib
import json
import pathlib
import shutil
import sys
import zipfile
from contextlib import ContextDecorator
from io import StringIO

import ansible
import ansible.modules
from ansible.module_utils import basic

import ansiblecall.utils.cache
import ansiblecall.utils.loader
from ansiblecall.utils.config import get_config
from ansiblecall.utils.respawn import respawn_module


class Context(ContextDecorator):
    """Run ansible module with certain sys methods overridden"""

    def __init__(self, module, params=None, runtime=None) -> None:
        super().__init__()

        self.__stdout = None
        self.__argv = None
        self.__path = None
        self.__ret = None

        # Store context inputs
        self.params = params or {}
        self.module = module
        self.runtime = runtime

    def cache(self, dest=None):
        return ansiblecall.utils.cache.cache(mod_name=self.module.key, dest=dest)

    def run(self):
        try:
            if self.runtime:
                ansible.module_utils.common.respawn.respawn_module(runtime=self.runtime)
            else:
                mod = importlib.import_module(self.module.name)
                mod.main()
        except SystemExit:
            return self.ret

    def __enter__(self):
        """Patch necessary methods to run an Ansible module"""
        self.__ret = StringIO()
        self.__stdout = sys.stdout
        self.__argv = sys.argv
        self.__path = sys.path

        # Patch ANSIBLE_ARGS. All Ansible modules read their parameters from
        # this variable.
        basic._ANSIBLE_ARGS = json.dumps(  # noqa: SLF001
            {"ANSIBLE_MODULE_ARGS": self.params or {}},
        ).encode("utf-8")

        # Patch respawn module
        ansible.module_utils.common.respawn.respawn_module = respawn_module

        # Patch sys module. Ansible modules will use sys.exit(x) to return
        sys.argv = []
        sys.stdout = self.__ret
        sys.modules["__main__"]._module_fqn = self.module.name  # noqa: SLF001
        sys.modules["__main__"]._modlib_path = self.module.path  # noqa: SLF001
        sys.modules["__main__"]._module_abs = self.module.abs  # noqa: SLF001
        if self.module.path not in sys.path:
            sys.path.insert(0, self.module.path)
        return self

    @staticmethod
    def clean_return(val):
        """All ansible modules print the return json to stdout.
        Read the return json in stdout from our StringIO object.
        """
        ret = None
        try:
            if val:
                val = val.strip().split("\n")[-1]
            ret = json.loads((val or "{}").strip())
            if "invocation" in ret:
                ret.pop("invocation")
        except (json.JSONDecodeError, TypeError) as exc:
            ret = str(exc)
        return ret

    @property
    def ret(self):
        """Grab return from stdout"""
        return self.clean_return(self.__ret.getvalue())

    def __exit__(self, *exc):
        """Restore all patched objects"""
        sys.argv = self.__argv
        sys.stdout = self.__stdout
        sys.path = self.__path
        self.__ret = None
        delattr(sys.modules["__main__"], "_module_fqn")
        delattr(sys.modules["__main__"], "_modlib_path")
        delattr(sys.modules["__main__"], "_module_abs")


class ZipContext(ContextDecorator):
    def __init__(self, mod_name):
        super().__init__()
        self.mod_name = mod_name

    def reload(self):
        import ansible

        zip_filename = pathlib.Path(ansible.__file__).parent.parent
        if not zipfile.is_zipfile(zip_filename):
            return
        target_dir = pathlib.Path(get_config(key="cache_dir")).joinpath(self.mod_name)
        if ansiblecall.utils.cache.compare_checksum(filename=zip_filename) is False or target_dir.exists() is False:
            if target_dir.exists():
                shutil.rmtree(target_dir)
            pathlib.Path(target_dir).mkdir(parents=True, exist_ok=True)
            with zipfile.ZipFile(file=zip_filename) as zp:
                zp.extractall(path=target_dir)
        sys.path.insert(0, str(target_dir))
        ansiblecall.utils.loader.load_mods.cache_clear()
        # Trigger re-import
        ansiblecall.utils.loader.reload()

    def __enter__(self):
        self.__path = sys.path
        self.reload()
        return self

    def __exit__(self, *exc):
        sys.path = self.__path



import os


class Config(dict):
    def __init__(self):
        super().__init__()
        self["cache_dir"] = os.path.expanduser(os.path.join("~", ".ansiblecall", "cache"))
        self["log_level"] = "info"

    def __getattr__(self, key):
        return self.get(key)

    def __setattr__(self, key, value):
        self[key] = value


def get_config(key=None):
    config = Config()
    if key:
        return config.get(key)
    return config



import hashlib
import logging
import os
import pathlib
import shutil
import sys
import tempfile

import ansiblecall.utils.config
import ansiblecall.utils.loader

CACHE_DIR = ansiblecall.utils.config.get_config(key="cache_dir")
if not os.path.exists(CACHE_DIR):
    os.makedirs(CACHE_DIR, exist_ok=True)

log = logging.getLogger(__name__)


def package_libs(path):
    """
    Package ansible module and util libraries at a given path
    """
    # Lazy import
    import ansible
    import ansible._vendor
    import ansible.module_utils
    import ansible.release

    import ansiblecall

    roots = {
        "ref": {
            "site_packages": pathlib.Path(ansible.__file__).parent.parent,
            "collections_root": pathlib.Path(sys.modules["__main__"]._modlib_path),  # noqa: SLF001
            "collections_plugins": pathlib.Path(
                sys.modules["__main__"]._module_abs  # noqa: SLF001
            ).parent.parent,
            "ansiblecall_root": pathlib.Path(ansiblecall.__file__).parent.parent,
        },
        "builtins": [
            {
                "src": pathlib.Path(ansible.__file__),
                "relative_to": "site_packages",
                "copytree": False,
            },
            {
                "src": pathlib.Path(ansible.module_utils.__file__).parent,
                "relative_to": "site_packages",
                "copytree": True,
            },
            {
                "src": pathlib.Path(ansible._vendor.__file__).parent,  # noqa: SLF001
                "relative_to": "site_packages",
                "copytree": True,
            },
            {
                "src": pathlib.Path(ansible.release.__file__),
                "relative_to": "site_packages",
                "copytree": False,
            },
            {
                "src": pathlib.Path(sys.modules["__main__"]._module_abs),  # noqa: SLF001
                "relative_to": "site_packages",
                "copytree": False,
            },
        ],
        "collections": [
            {
                "src": pathlib.Path(sys.modules["__main__"]._module_abs),  # noqa: SLF001
                "relative_to": "collections_root",
                "copytree": False,
            },
            {
                "src": "collections_plugins",
                "joinpath": "module_utils",
                "relative_to": "collections_root",
                "copytree": True,
            },
            {
                "src": "collections_plugins",
                "joinpath": "plugin_utils",
                "relative_to": "collections_root",
                "copytree": True,
            },
        ],
        "ansiblecall": [
            {
                "src": pathlib.Path(ansiblecall.__file__).parent,
                "relative_to": "ansiblecall_root",
                "copytree": True,
            }
        ],
    }
    module_fqdn = sys.modules["__main__"]._module_fqn  # noqa: SLF001
    ref, sources = roots["ref"], roots["builtins"]
    if not module_fqdn.startswith("ansible.modules."):
        sources += roots["collections"]
    sources += roots["ansiblecall"]
    for s in sources:
        src = ref[s["src"]] if isinstance(s["src"], str) else s["src"]
        joinpath = s.get("joinpath")
        if joinpath:
            src = src.joinpath(joinpath)
        dst = pathlib.Path(path).joinpath(src.relative_to(ref[s["relative_to"]]))
        copytree = s["copytree"]
        dirs = dst if copytree else dst.parent
        if not src.exists():
            continue
        os.makedirs(dirs, exist_ok=True)
        if copytree:
            shutil.copytree(
                src=src,
                dst=dst,
                dirs_exist_ok=True,
                ignore=shutil.ignore_patterns("__pycache__", "*.pyc"),
            )
        else:
            shutil.copy(src=src, dst=dst)


def compare_checksum(filename: str):
    zip_name = str(filename).removesuffix(".zip")
    checksum = ""
    checksum_file = zip_name + ".sha256"
    if os.path.exists(checksum_file):
        with open(checksum_file) as hfp:
            checksum = hfp.read()
    zip_checksum = get_checksum(filename=filename)
    return checksum == zip_checksum


def get_checksum(filename: str):
    zip_name = str(filename).removesuffix(".zip")
    with open(zip_name + ".zip", "rb") as fp:
        return hashlib.sha256(fp.read()).hexdigest()


def save_checksum(filename: str):
    checksum = get_checksum(filename=filename)
    zip_name = str(filename).removesuffix(".zip")
    with open(zip_name + ".sha256", "w") as hfp:
        hfp.write(checksum)
    return checksum


def cache(mod_name, dest=None):
    with tempfile.TemporaryDirectory() as tmp_dir:
        package_libs(path=tmp_dir)
        archive_name = os.path.join(CACHE_DIR, mod_name)
        shutil.make_archive(archive_name, format="zip", root_dir=tmp_dir)
        checksum = save_checksum(filename=archive_name)
        if dest:
            for ext in [".zip", ".sha256"]:
                pathlib.Path(dest).mkdir(parents=True, exist_ok=True)
                dest_file = pathlib.Path(dest).joinpath(mod_name + ext)
                if dest_file.exists():
                    dest_file.unlink(missing_ok=True)
                shutil.move(src=archive_name + ext, dst=dest)
        log.debug("Cached %s module at %s.", mod_name, dest or CACHE_DIR)
        return checksum


def refresh_modules():
    """Refresh Ansible module cache"""
    fun = ansiblecall.utils.loader.load_mods
    fun.cache_clear()
    return fun()



import hashlib
import os
import pathlib
import shutil
import tempfile

import pytest

import ansiblecall
import ansiblecall.utils.loader

IS_ROOT = os.getuid() == 0


def not_debian():
    if os.path.exists("/etc/os-release"):
        with open("/etc/os-release") as f:
            os_release_info = f.read().lower()
            if "debian" in os_release_info or "ubuntu" in os_release_info or "mint" in os_release_info:
                return False
    return True


NOT_DEBIAN = not_debian()


def test_ansiblecall_module():
    """Ensure ansible module can be called as an ansiblecall module"""
    assert ansiblecall.module(mod_name="ansible.builtin.ping", data="hello") == {"ping": "hello"}
    assert ansiblecall.module(mod_name="ansible.builtin.ping") == {"ping": "pong"}
    with tempfile.TemporaryDirectory() as tmp_dir:
        foo_file = str(pathlib.Path(tmp_dir).joinpath("foo"))
        foo_archive = str(pathlib.Path(tmp_dir).joinpath("foo.gz"))

        ret = ansiblecall.module(mod_name="ansible.builtin.file", path=foo_file, state="absent")

        ret = ansiblecall.module(mod_name="ansible.builtin.file", path=foo_file, state="touch")
        assert ret["changed"] is True
        ansiblecall.module(mod_name="ansible.builtin.file", path=foo_archive, state="absent")
        ret = ansiblecall.module(mod_name="community.general.archive", path=foo_file)
        assert ret["changed"] is True
        ret = ansiblecall.module(mod_name="community.general.archive", path=foo_file)
        assert ret["changed"] is False


def test_module_refresh():
    """Ensure modules are refreshed"""
    assert ansiblecall.refresh_modules()


@pytest.mark.skipif(NOT_DEBIAN or not IS_ROOT, reason="Not debian distro, or non-root user")
def test_respawn_root_user():
    """Ensure ansible modules like apt which use respawn works"""
    assert ansiblecall.module(mod_name="ansible.builtin.ping") == {"ping": "pong"}
    # Install hello package
    ret = ansiblecall.module(mod_name="ansible.builtin.apt", name="hello", state="absent")
    ret = ansiblecall.module(mod_name="ansible.builtin.apt", name="hello", state="present")
    assert ret["changed"] is True
    ret = ansiblecall.module(mod_name="ansible.builtin.apt", name="hello", state="present")
    assert ret["changed"] is False


def test_cache():
    """Check if cache can be created and verify the sum"""
    shutil.rmtree(os.path.expanduser("~/.ansiblecall"))
    hash_sum = ansiblecall.cache(mod_name="ansible.builtin.file")
    with open(os.path.expanduser("~/.ansiblecall/cache/ansible.builtin.file.zip"), "rb") as fp:
        expected = hashlib.sha256(fp.read()).hexdigest()
    assert hash_sum == expected


def test_cache_dest():
    """Ensure modules are cacheable to a specified dir"""
    with tempfile.TemporaryDirectory() as tmp_dir:
        mod_name = "ansible.builtin.ping"
        ansiblecall.cache(mod_name=mod_name, dest=tmp_dir)
        for ext in [".zip", ".sha256"]:
            assert pathlib.Path(tmp_dir).joinpath(mod_name + ext).exists()



import pathlib
import tempfile

import pytest

from ansiblecall.utils import typefactory


@pytest.mark.timeout(10)
def test_type_factory():
    """Ensure ansible module can be called using typings"""
    typefactory.TypeFactory.run(
        modules=[
            "ansible.builtin.ping",
            "community.general.archive",
            "ansible.builtin.file",
        ],
        clean=True,
    )
    import ansiblecall.typed.ansible_builtin_file as file
    import ansiblecall.typed.ansible_builtin_ping as ping
    import ansiblecall.typed.community_general_archive as archive

    ret = ping.Ping(data="hello").run()
    assert ret.ping == "hello"
    ret = ping.Ping(data="hello").raw()
    assert ret == {"ping": "hello"}
    p = ping.Ping()
    p.data = "ping"
    assert p.run().ping == "ping"

    with tempfile.TemporaryDirectory() as tmp_dir:
        foo_file = str(pathlib.Path(tmp_dir).joinpath("foo"))
        foo_archive = str(pathlib.Path(tmp_dir).joinpath("foo.gz"))
        ret = file.File(path=foo_file, state="absent").run()
        ret = file.File(path=foo_file, state="touch").run()
        assert ret.changed is True
        ret = file.File(path=foo_archive, state="absent").run()
        ret = archive.Archive(path=foo_file).run()
        assert ret.changed is True
        ret = archive.Archive(path=foo_file).run()
        assert ret.changed is False



import logging
import time

import ansiblecall.utils.cache
import ansiblecall.utils.config
import ansiblecall.utils.ctx
import ansiblecall.utils.loader
from ansiblecall.utils.rt import Runtime

log = logging.getLogger(__name__)

if not log.hasHandlers():
    logging.basicConfig(
        level=getattr(logging, ansiblecall.utils.config.get_config().log_level.upper()),
        format="%(asctime)s [%(name)-17s][%(levelname)-8s:%(lineno)-4d][%(processName)s:%(process)d] %(message)s",
    )


def module(mod_name, *, rt: Runtime = None, **params):
    """Run ansible module."""
    start = time.time()
    log.debug("Running module [%s] with params [%s]", mod_name, ", ".join(list(params)))
    mod = ansiblecall.utils.loader.get_module(mod_name=mod_name)
    with ansiblecall.utils.ctx.Context(module=mod, params=params, runtime=rt) as ctx:
        ret = ctx.run()
        log.debug(
            "Returning data to caller. Total Elapsed: %0.03fs",
            (time.time() - start),
        )

        return ret


def refresh_modules():
    """Refresh Ansible module cache"""
    return ansiblecall.utils.cache.refresh_modules()


def cache(mod_name, dest=None):
    """Cache ansible modules and dependencies into a zip file"""
    mod = ansiblecall.utils.loader.get_module(mod_name=mod_name)
    with ansiblecall.utils.ctx.Context(module=mod) as ctx:
        return ctx.cache(dest=dest)


def config():
    """Get configuration parameters"""
    return ansiblecall.utils.config.get_config()



class Runtime(dict):
    def __init__(self, *, become=False, become_user=""):
        super().__init__()
        self.become = become
        self.become_user = become_user

    def __getattr__(self, key):
        return self.get(key)

    def __setattr__(self, key, value):
        self[key] = value



import ast
import contextlib
import dataclasses
import json
import logging
import multiprocessing
import pathlib
import shutil

import yaml

import ansiblecall
import ansiblecall.utils.loader

log = logging.getLogger(__name__)


@dataclasses.dataclass(kw_only=True)
class InputBase:
    rt: ansiblecall.Runtime = None


@dataclasses.dataclass(kw_only=True)
class OutputBase:
    failed: bool = None
    msg: str = None
    rc: int = None
    changed: bool = None
    diff: dict = None
    skipped: bool = None
    backup_file: str = None
    results: list = None
    stderr: str = None
    stderr_lines: list = None
    stdout: str = None
    stdout_lines: list = None


@dataclasses.dataclass(kw_only=True)
class Field:
    name: str
    optional: bool
    type: str
    default: str
    description: str
    choices: list[str]
    elements: str

    def format_default(self):
        ret = None
        if self.type is bool and isinstance(self.default, str) and self.default is not None:
            if self.default.lower() in ["yes", "true"]:
                ret = True
            elif self.default.lower() in ["no", "false"]:
                ret = False
        elif self.type is str and self.default is not None:
            ret = f"{self.default!r}"
        elif self.type is dict and isinstance(self.default, str) and self.default is not None:
            ret = json.loads(self.default)
        elif (self.type is float or self.type is int) and self.default is not None:
            ret = self.type(self.default)
        return ret

    def __repr__(self):
        default = f"= {self.format_default()}" if self.optional else ""
        description = " ".join(self.description) if isinstance(self.description, list) else self.description
        choices = (self.choices and "Choices: " + str(self.choices)) or ""
        return f'{self.name}: {self.type.__name__} {default}\n"""{description} {choices}"""'


class TypeFactory:
    def __init__(
        self,
        type_dir: str,
        module_name: dict,
    ):
        self.type_dir = type_dir
        self.module_name = module_name

        self.output_class_name = None
        self.output_class_body = None
        self.input_class_name = None
        self.input_class_body = None
        self.module_file_name = None

    @staticmethod
    def convert_fields_to_lines(fields):
        ret = []
        for f in fields:
            lines = str(f).split("\n")
            ret.extend([line.strip() for line in lines])
        return ret

    @staticmethod
    def align(lines):
        ret = ""
        for line in lines:
            ret += f"    {line}\n"
        return ret

    @classmethod
    def generate_class_body(cls, fields):
        lines = cls.convert_fields_to_lines(fields=fields)
        return cls.align(lines)

    def generate(self):
        mods = ansiblecall.utils.loader.load_mods()
        schema = self.get_io_schema(mod=mods[self.module_name])
        self.input_class_name = self.module_name.split(".")[2].capitalize()
        self.output_class_name = f"{self.input_class_name}Out"
        self.output_class_body = self.generate_class_body(fields=schema["output"])
        self.input_class_body = self.generate_class_body(fields=schema["input"])
        self.module_file_name = f"{self.module_name.replace('.', '_')}.py"
        code = self.render_template()
        with open(pathlib.Path(self.type_dir).joinpath(self.module_file_name), "w") as fp:
            fp.write(code)

    def render_template(self):
        return f"""import dataclasses
import ansiblecall
import ansiblecall.utils.typefactory


@dataclasses.dataclass(kw_only=True)
class {self.output_class_name}(ansiblecall.utils.typefactory.OutputBase):
{self.output_class_body}
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

@dataclasses.dataclass(kw_only=True)
class {self.input_class_name}(ansiblecall.utils.typefactory.InputBase):
{self.input_class_body}
    # Method to filter out unset values or values left at their default
    def get_params(self) -> dict:
        ret = {{}}
        for f in dataclasses.fields(self):
            value = getattr(self, f.name)
            # Check if the value is different from the default (or explicitly set)
            if value is not None and value != f.default:
                ret[f.name] = value
        return ret

    def run(self) -> {self.output_class_name}:
        return {self.output_class_name}(**self.raw())

    def raw(self) -> dict:
        return ansiblecall.module({self.module_name!r}, **self.get_params())
"""

    @classmethod
    def process(cls, queue):
        """
        Run type generation using multiprocessing lib
        """
        num_procs = multiprocessing.cpu_count()
        with multiprocessing.Pool(num_procs) as p:
            p.map(cls.generate_parallel, [queue] * num_procs)

    @staticmethod
    def get_var_value(mod_str: str, var: str) -> str:
        """
        Return value of a variable in a python module
        """
        return next(
            (
                n.value.value
                for n in ast.walk(ast.parse(mod_str))
                if isinstance(n, ast.Assign) and hasattr(n.targets[0], "id") and n.targets[0].id == var
            ),
            None,
        )

    @staticmethod
    def parse_yaml(doc: str) -> dict:
        """
        Parse doc yaml
        """
        ret = {}
        with contextlib.suppress(yaml.YAMLError):
            ret = yaml.safe_load(doc) or {}
        return ret

    @staticmethod
    def parse_fragment(fragments):
        """
        Parse a doc fragment into a field schema
        """
        ret = []
        type_map = {
            "dict": dict,
            "int": int,
            "path": str,
            "str": str,
            "any": str,
            "sid": str,
            "float": float,
            "bool": bool,
            "jsonarg": str,
            "complex": dict,
            "json": str,
            "raw": str,
            "list": list,
            None: str,
        }
        for name, fragment in fragments.items():
            if not isinstance(fragment, dict):
                continue
            type_ = type_map[fragment.get("type", "str")]
            elements = fragment.get("elements", "")
            optional = not (fragment.get("always") or fragment.get("required"))
            default = fragment.get("default")
            description = fragment.get("description", "")
            choices = fragment.get("choices")
            ret.append(
                Field(
                    name=name,
                    optional=optional,
                    elements=elements,
                    type=type_,
                    default=default,
                    description=description,
                    choices=choices,
                )
            )
        return ret

    @classmethod
    def get_io_schema(cls, mod: dict) -> dict[str, str]:
        """
        Get input and output docs for a module
        """
        ret, mod_str = {}, ""
        with open(mod.abs) as fp:
            mod_str = fp.read()
        for doc_var, var in (("DOCUMENTATION", "input"), ("RETURN", "output")):
            val = cls.get_var_value(mod_str=mod_str, var=doc_var)
            parsed = {}
            if val:
                parsed = cls.parse_yaml(val)
            fragments = parsed.get("options") if "options" in parsed else parsed
            ret[var] = cls.parse_fragment(fragments=fragments)
        return ret

    @staticmethod
    def init_dirs(clean=None):
        init_file = pathlib.Path(__file__).parent.parent.joinpath("typed", "__init__.py")
        init_dir = init_file.parent
        if clean and init_dir.exists():
            log.info("Removing dir %s.", init_dir)
            shutil.rmtree(init_dir)
        init_dir.mkdir(parents=True, exist_ok=True)
        log.info("Created dir %s.", init_dir)
        init_file.touch()
        return init_dir

    @staticmethod
    def generate_parallel(queue):
        while not queue.empty():
            factory = queue.get()
            factory.generate()
            with contextlib.suppress(NotImplementedError):
                queue_size = queue.qsize()
                if queue_size and queue_size % 500 == 0:
                    log.info("%s modules remaining.", queue_size)

    @classmethod
    def run(cls, modules=None, clean=None):
        """
        Install typings for ansible modules
        """
        mods = ansiblecall.refresh_modules()
        type_mods = (modules and list(set(modules) & set(mods))) or list(mods)
        if clean is None and not modules:
            clean = True
        log.info("Initializing dirs.")
        type_dir = cls.init_dirs(clean=clean)
        log.info("Generating types for %s module(s).", len(type_mods))
        multiprocessing.set_start_method("spawn")
        with multiprocessing.Manager() as m:
            queue = m.Queue()
            for module_name in type_mods:
                factory = cls(
                    type_dir=type_dir,
                    module_name=module_name,
                )
                queue.put(factory, block=False)
            cls.process(queue=queue)
            log.info("Done!")
